#!/usr/bin/env cdat
"""
A set of functions for reading in and processing CMIP6 data
"""

#IMPORT STUFF:
#=====================
import MV2 as MV
import numpy as np
import gc
import PMC_utils
import MDZ_utils as MU


###########################################################################
def get_plottable_time(X):
    # Function stolen / modified from Kate
    years = [x.year+(x.month-1)/12. for x in X.getTime().asComponentTime()]
    months = [x.month for x in X.getTime().asComponentTime()]
    return (np.array(years),np.array(months))


###########################################################################
def get_pressure_weights2new(pres,ps,tropp):

    # much faster version -- no loops!

    # put tropp in Pa.
    ptrop=100*tropp
    # make sure pres is in Pa.
    if MV.maximum(pres)==1000:
        pres=100*pres   
    # make sure ps is in Pa.
    if MV.maximum(ps)<2000:
        ps=100*ps

    #CONSTRUCT LAYER PRESSURE DIFFS:
    #==============================
    # get pressure level diffs in Pa:
    dp=pres[0:-1]-pres[1:]
    
    a,b,c=ps.shape
    trop_wts=MV.resize(dp,[a,b,c,len(pres)-1])(order=[0,3,1,2])
    
    atm_wts=trop_wts.clone() # clone seems to also copy the AxisList

    # SET EVERYTHING ABOVE THE TROPOPAUSE TO 0  
    rep_ptrop0 = np.broadcast_to(ptrop,(17,a,b,c),subok=True)
    rep_ptrop = MV.array(rep_ptrop0)(order=[1,0,2,3])
    rep_pres0 = np.broadcast_to(pres,(a,b,c,17),subok=True)
    rep_pres = MV.array(rep_pres0)(order=[0,3,1,2])
    
    # Compute local minus tropopause pressure, which should be:
    # negative above the trop
    # smaller than trop_wts just below trop (i.e., this pressure is closer to the ptrop than to the plev above)
    # larger that trop_wts well below trop (i.e., this pressure is closer the plev above than to the ptrop)
    p_rel_ptrop = rep_pres - rep_ptrop    
    
    # trop_wts is the smaller of dp and p_rel_ptrop in the troposphere:
    trop_wts = MV.where(trop_wts>p_rel_ptrop[:,:-1,:],p_rel_ptrop[:,:-1,:],trop_wts)
    
    # trop_wts is zero in the stratosphere:
    trop_wts = MV.where(p_rel_ptrop[:,:-1,:]<0,0,trop_wts)
    
    
    # SET EVERYTHING BELOW THE SURFACE TO 0
    rep_ps0 = np.broadcast_to(ps,(17,a,b,c),subok=True)
    rep_ps = MV.array(rep_ps0)(order=[1,0,2,3])    
    
    # Compute local minus surface pressure, which should be:
    # positive for pressures below the sfc
    # smaller than trop_wts just above the sfc (i.e., this pressure is closer to the sfc than to the plev above)
    # larger that trop_wts well above the sfc (i.e., this pressure is closer to the plev above than to the sfc)
    p_rel_sfc = rep_pres - rep_ps
    
    # trop_wts is the smaller of dp and abs(p_rel_sfc) in the troposphere:
    trop_wts = MV.where(trop_wts>abs(p_rel_sfc[:,1:,:]),abs(p_rel_sfc[:,1:,:]),trop_wts)
    atm_wts = MV.where(atm_wts>abs(p_rel_sfc[:,1:,:]),abs(p_rel_sfc[:,1:,:]),atm_wts)
    
    # trop_wts is zero under ground:
    trop_wts = MV.where(p_rel_sfc[:,1:,:]>=0,0,trop_wts)     
    atm_wts = MV.where(p_rel_sfc[:,1:,:]>=0,0,atm_wts)     
    
    # divide by 100 twice to put this in hPa and to weight the layer relative to perturbations in 100 hPa layers
    trop_wts=trop_wts/10000. 
    atm_wts=atm_wts/10000.        

    return (trop_wts,atm_wts)
    
###########################################################################
def get_weights_SPC(ps, trop, field):
    """
    from Stephen Po-Chedley
    
    wts = get_weights(ps, trop, field)

    Function returns the layer weights (layer thickness) given the upper
    integration boundary (e.g., the tropopause, trop) and the lower bound
    (e.g., surface pressure, ps). Uses the input field to get pressure level
    information.
                    ps - response field [time, lat, lon]
                    trop - radiative kernel [time, lat, lon]
                    field - field that will be integrated [time, plev, lat, lon]

    """     
    
            
    PAL = MV.zeros(ps.shape)
    PAT = MV.ones(ps.shape)          
    trop_wts = MV.zeros(field.shape - MV.array([0,1,0,0]))
    atm_wts = MV.zeros(field.shape - MV.array([0,1,0,0]))
    plev = field.getLevel()[:]
    
    # make sure pressures are in Pa.  
    if MV.maximum(plev)<=2000:
        plev=100*plev         
    if MV.maximum(ps)<=2000:
        ps=100*ps
    if MV.maximum(trop)<=2000:
        trop=100*trop
            
    if plev[0] < plev[1]:
        raise ValueError('This script assumes that plev[0] > plev[1].')
    for i in range(len(plev)-1):
        # allocate slice
        sfield = MV.zeros(ps.shape)
        tfield = MV.zeros(ps.shape)
        # get first level above surface
        p = plev[i]
        pp = plev[i+1]
        ISURF = MV.greater(ps, p) # 1 where ps>p, 0 where ps<p
        II = MV.equal(PAL, 0)
        IA = MV.logical_and(ISURF, II)                  
        PAL = MV.where(IA, pp, PAL)
        # get first layer below tropopause
        ITROP = MV.less(trop, p)
        II = MV.greater_equal(trop, pp)
        IB = MV.logical_and(ITROP, II)
        PAT = MV.where(IB, p, PAT)
        # layers above tropopause or below surface (i.e., zero weight)
        above_trop = MV.logical_not(ITROP)
        below_surf = MV.logical_not(ISURF)
        IC = MV.logical_or(below_surf,above_trop)
        # layers adjacent to both tropopause and surface (i.e., layer is between tropopause and surface)
        ID = MV.logical_and(IA, IB)
        # layers not adjacent to tropopause or surface (i.e., full layers)
        IE = MV.logical_or(IA, IB)
        IE = MV.logical_and(MV.logical_not(IC), MV.logical_not(IE)) 
        # layers not adjacent to surface (i.e., full layers)
        IF = MV.logical_and(MV.logical_not(below_surf), MV.logical_not(IA)) 
        # TROPOSPHERIC VALUES
        sfield = MV.where(IA, ps-PAL, sfield)
        sfield = MV.where(IB, PAT-trop, sfield)
        sfield = MV.where(IC, 0., sfield)
        sfield = MV.where(ID, ps-trop, sfield)
        sfield = MV.where(IE, p - pp, sfield)
        # store field and weight by per 100 hPa (1/100 for Pa to hPa and 1/100 for per *100* hPa)
        trop_wts[:, i, :, :] = sfield / 10000.
        
        # ATMOSPHERIC VALUES
        tfield = MV.where(IA, ps-PAL, tfield)
        tfield = MV.where(below_surf, 0., tfield)
        tfield = MV.where(IF, p - pp, tfield)
        # store field and weight by per 100 hPa (1/100 for Pa to hPa and 1/100 for per *100* hPa)
        atm_wts[:, i, :, :] = tfield / 10000.
        
    return trop_wts,atm_wts

###########################################################################
def get_pressure_weights3new(pres,ps,ptrop):

    # same as get_pressure_weights2new, but uses the weights other people use (e.g., Kramer)
    
    # ptrop is in hPa.
    # make sure pres is in hPa.
    if MV.maximum(pres)==100000:
        pres=pres/100.
    # make sure ps is in hPa.
    if MV.maximum(ps)>2000:
        ps=ps/100.

    #CONSTRUCT LAYER PRESSURE DIFFS:
    #==============================
    # These are taken from Ryan Kramer's script:
    #Pressure of upper boundary of each vertical layer
    pt = np.array([962.5, 887.5, 775, 650, 550, 450, 350, 275, 255, 175, 125, 85, 60, 40, 25, 15, 0])
    #Pressure of lower boundary of each vertical layer
    pb = np.array([1000, 962.5, 887.5, 775, 650, 550, 450, 350, 275, 225, 175, 125, 85, 60, 40, 25, 15])
    #Pressure thickness of each vertical level
    dp = np.array([37.5, 75, 112.5, 125, 100, 100, 100, 75, 50, 50, 50, 40, 25, 20, 15, 10, 15])
        
   
    a,b,c=ps.shape
    trop_wts=MV.resize(dp,[a,b,c,len(pres)])(order=[0,3,1,2])
    
    atm_wts=trop_wts.clone() # clone seems to also copy the AxisList

    ##########################################
    # SET EVERYTHING ABOVE THE TROPOPAUSE TO 0
    ##########################################  
    rep_ptrop0 = np.broadcast_to(ptrop,(17,a,b,c),subok=True)
    rep_ptrop = MV.array(rep_ptrop0)(order=[1,0,2,3])
    rep_pt0 = np.broadcast_to(pt,(a,b,c,17),subok=True)
    rep_pt = MV.array(rep_pt0)(order=[0,3,1,2])
    rep_pb0 = np.broadcast_to(pb,(a,b,c,17),subok=True)
    rep_pb = MV.array(rep_pb0)(order=[0,3,1,2])
    
    # Compute p_lower_boundary minus tropopause pressure, which should be:
    # negative above the trop
    # smaller than trop_wts just below trop (i.e., this pressure is closer to the ptrop than to the plev above)
    # larger that trop_wts well below trop (i.e., this pressure is closer the plev above than to the ptrop)
    pb_rel_ptrop = rep_pb - rep_ptrop    
    
    # trop_wts is the smaller of dp and pb_rel_ptrop in the troposphere:
    trop_wts = MV.where(trop_wts>pb_rel_ptrop,pb_rel_ptrop,trop_wts)
    
    # trop_wts is zero in the stratosphere:
    trop_wts = MV.where(pb_rel_ptrop<0,0,trop_wts)
    
    
    ##########################################
    # SET EVERYTHING BELOW THE SURFACE TO 0
    ##########################################
    rep_ps0 = np.broadcast_to(ps,(17,a,b,c),subok=True)
    rep_ps = MV.array(rep_ps0)(order=[1,0,2,3])    
    
    # Compute p_upper_boundary minus surface pressure, which should be:
    # positive for pressures below the sfc
    # smaller than trop_wts just above the sfc (i.e., this pressure is closer to the sfc than to the plev below)
    # larger than trop_wts well above the sfc (i.e., this pressure is closer the plev below than to the sfc)
    pt_rel_sfc = rep_pt - rep_ps
    
    # trop_wts is the smaller of dp and abs(pt_rel_sfc) in the troposphere:
    trop_wts = MV.where(trop_wts>abs(pt_rel_sfc),abs(pt_rel_sfc),trop_wts)
    atm_wts = MV.where(atm_wts>abs(pt_rel_sfc),abs(pt_rel_sfc),atm_wts)
    
    # trop_wts is zero under ground:
    trop_wts = MV.where(pt_rel_sfc>=0,0,trop_wts)     
    atm_wts = MV.where(pt_rel_sfc>=0,0,atm_wts) 
    
        
    
    # divide by 100 to weight the layer relative to perturbations in 100 hPa layers
    trop_wts=trop_wts/100. 
    atm_wts=atm_wts/100.        

    return (trop_wts,atm_wts)
    
   



###########################################################################
def get_tropopause_pressure(kern_plev,TA):

    # Perform tropopause pressure calculation
    # generated via: f2py -c -m tropo tropo.f90 
    # tropo.f90 came from http://www.inscc.utah.edu/~reichler/research/projects/TROPO/code.txt

    import tropo  

    PP=kern_plev/100.
    plimu=45000 # upper limit for tropopause pressure in Pa, usually 45000
    pliml=7500 # lower limit for tropopause pressure in Pa, usually 7500
    plimlex=7500 # lower limit in extratropics, usually same as pliml, i.e., 7500
    dofill=1 # 1 = fill undefined values with neighboring points if 1, 0 = do not fill
    # fortran wants this in lon,lat,ctp but lets swap time with lon
    tropp=MV.zeros((TA[:,0,:].shape))
    tropp=MV.masked_where(tropp==0,tropp)
    LO = TA.shape[-1]
    for LO in range(LO):
        temp=TA[:,:,:,LO](order=[0,2,1]) 
        tp,tperr=tropo.tropo(temp,PP,plimu,pliml,plimlex,dofill)

        tropp[:,:,LO]=tp/100. # convert to hPa

    tropp=MV.masked_where(tropp>plimu/100.,tropp)
    tropp=MV.masked_where(tropp<0.,tropp)

    return tropp

    
###########################################################################
def nanarray(vector):
    """
    this generates a masked array with the size given by vector
    example: vector = (90,144,28)
    similar to this=NaN*ones(x,y,z) in matlab
    """

    this=MV.zeros(vector)
    this=MV.masked_where(this==0,this)

    return this

###########################################################################
def running_mean(data,N):

    # Compute N-yr running mean
    back = N/2
    forward = N/2+1
    avgdata=MU.nanarray(data.shape)
    for month in range(12):
        subset = data[month::12,:]
        avgsubset = MU.nanarray(subset.shape)
        LS=len(subset)
        for ii in range(LS):
            start = np.max((ii-back,0))
            finish = np.min((ii+forward,LS))
            avgsubset[ii,:] = MV.average(subset[start:finish,:],0)
        avgdata[month::12,:] = avgsubset
        
    avgdata.setAxisList(data.getAxisList())
    
    return avgdata


###########################################################################

def ta_normalized_qv(kern_plev,T0,T1,q0,q1,q_norm_flag='',blend_flag='',lognorm_flag='angie'):
    
    #this function is modified from Peter's get_alb function
    # returns an array of delta hus values that equal the amount necessary 
    # to keep RH fixed if the 21 yr running mean-smoothed piControl 
    # temperature is perturbed by 1 K            

    if q_norm_flag=='':
        avgta = (T0 + T1)/2. # we used to simply use T0 here
    elif q_norm_flag=='q_norm_piT':
        avgta = T0
    elif q_norm_flag=='q_norm_abT':
        avgta = T1

    #CALCULATE SAT SPECIFIC HUMIDITY AT T AND T+1
    #=========================================
    #qsat_water needs a pressure matrix of same dims as T0=[time,height,lat,lon] but we only have plev 
    #with dims=[height]. The resize command just copies plev a bunch of times to fit into a 1-d array of 
    #length dim0*dim1*dim2*dim3 then maps your [i,j,k,l] indexing to this 1-d array according to 
    #l + dim(3)*k + dim(2)*dim(3)*j + dim(1)*dim(2)*dim(3)*i... in other words, plev is strided across the *LAST* 
    #dimension. Therefore it only makes sense to call resize with ht = the last dimension:
    pres=MV.resize(kern_plev,[T0.shape[0],T0.shape[2],T0.shape[3],T0.shape[1]])\
        (order=[0,3,1,2]) #0 was time, 3 was last=ht, and 1,2 were lat,lon...
    #     ^ but we wanted ht to be the 1st dim (in zero indexing) so we have to reorder...

    # PETER'S QSAT
    # use the avg of ctl and abrupt in computing Delq
    wsl=MU.qsat_water(avgta,pres)
    wsl_plus1=MU.qsat_water(avgta+1,pres)
    wsi=PMC_utils.qsat_ice(avgta,pres)
    wsi_plus1=PMC_utils.qsat_ice(avgta+1,pres)
    qsl=wsl/(1+wsl) # convert from mixing ratio (kg/kg) to specific humidity
    qsl_plus1=wsl_plus1/(1+wsl_plus1)
    qsi=wsi/(1+wsi) # convert from mixing ratio (kg/kg) to specific humidity
    qsi_plus1=wsi_plus1/(1+wsi_plus1)   


    if blend_flag=='':
        # Compute blend of qsi and qsl between -40 and 0
        blend = (avgta-233)*qsl/40 + (273-avgta)*qsi/40
        blend_plus1 = (avgta-233)*qsl_plus1/40 + (273-avgta)*qsi_plus1/40

        qs = np.where((avgta>233) & (avgta<273), blend, qsi)#[0]
        qs_plus1 = np.where((avgta>233) & (avgta<273), blend_plus1, qsi_plus1)#[0]
        
        qs = np.where(avgta >= 273, qsl, qs)#[0]
        qs_plus1 = np.where(avgta >= 273, qsl_plus1, qs_plus1)#[0]
        
        del blend, blend_plus1
        
    elif blend_flag=='T273':  # use qsat w.r.t. ice below 0C and and w.r.t. liquid above 0C      
        qs = np.where(avgta >= 273, qsl, qsi)#[0]
        qs_plus1 = np.where(avgta >= 273, qsl_plus1, qsi_plus1)#[0]
        
    elif blend_flag=='T233':  # use qsat w.r.t. ice below -40C and and w.r.t. liquid above -40C      
        qs = np.where(avgta >= 233, qsl, qsi)#[0]
        qs_plus1 = np.where(avgta >= 233, qsl_plus1, qsi_plus1)#[0]
        
    elif blend_flag=='wrtliq': # use qsat w.r.t. liquid everywhere      
        qs = qsl#[0]
        qs_plus1 = qsl_plus1#[0]
        
        
    if lognorm_flag=='angie':
        #=========================================
        # Angie's Logarithmic method; largest result because largest numerator and denominator only slightly larger
        #=========================================
        #1) Compute dlog(q) as dq/q0 [unitless]
        dq = q1 - q0
        dlogq = np.ma.true_divide(dq,q0) # larger than hus_anom = MV.log(q1) - MV.log(q0) 
        
        #2) Compute normalization term [per K] -- this is identical to Kramer's normalization
        dqsd1K = qs_plus1 - qs
        RH = np.ma.true_divide(q0,qs)
        qnorm = np.ma.true_divide(RH,q0)*dqsd1K
        #dlogqs = np.ma.true_divide(dqsd1K,qs) # identical to qnorm; larger than Delq = MV.log(qs_plus1) - MV.log(qs)
        
        #3) Normalize the humidity anomalies to put them in units of Kelvin:
        norm_hus_anom = np.ma.true_divide(dlogq,qnorm)
        norm_hus_anom.setAxisList(q1.getAxisList())
        
    elif lognorm_flag=='angie_linear':
        #=========================================
        # Angie's Linear method; identical to previous
        #=========================================
        #1) Compute dq [kg/kg]
        dq = q1 - q0
        
        #2) Compute normalization term [kg/kg per K]
        # Calculate the moisture change in response to 1 K warming at constant relative humidity
        rh = np.ma.true_divide(q0,qs)  # Relative humidity
        q1k = rh*qs_plus1              # New mixing ratio
        dq1k = q1k - q0                # Mixing ratio response for 1 K warming        
        
        #3) Normalize the humidity anomalies to put them in units of Kelvin:
        norm_hus_anom = np.ma.true_divide(dq,dq1k)
        norm_hus_anom.setAxisList(q1.getAxisList())
        
    elif lognorm_flag=='kramer':
        #=========================================
        # Ryan Kramer's method; smallest result because smallest numerator and largest denominator
        #=========================================
        #1) Compute dq
        dq = MV.log(q1) - MV.log(q0)
        
        #2) Compute normalization term -- this is identical to Angie's logarithmic normalization
        # Calculate the moisture change in response to 1 K warming at constant relative humidity
        qs_ratio_m1 = np.ma.true_divide(qs_plus1,qs) - 1
                
        #3) Normalize the humidity anomalies to put them in units of Kelvin:
        norm_hus_anom = np.ma.true_divide(dq,qs_ratio_m1)
        norm_hus_anom.setAxisList(q1.getAxisList())
        
    elif lognorm_flag=='mark_adj':
    
        #=========================================
        # Alternative Zelinka Method; 
        #=========================================
        
        #COMPUTE THE NORMALIZED ANOMALIES
        #=========================================
        Delq = np.ma.true_divide(qs_plus1-qs,qs)
       
        #GET CTL and FUT HUS AND COMPUTE DIFFERENCE OF LOGS 
        #===========================================================================    
        hus_anom = np.ma.true_divide(q1-q0,q0)

        # KERNELS REQUIRE DETRENDED QV DIVIDED BY DELQ:
        #=========================================
        norm_hus_anom = np.ma.true_divide(hus_anom,Delq)
        norm_hus_anom.setAxisList(q1.getAxisList())    
        del hus_anom, Delq

        
    else:
    
        #=========================================
        # Zelinka Method; smaller than Angie / bigger than Kramer b/c/ smallest numerator and only slightly smaller denominator
        #=========================================
        
        #COMPUTE THE NORMALIZED ANOMALIES
        #=========================================
        Delq = MV.log(qs_plus1) - MV.log(qs)
       
        #GET CTL and FUT HUS AND COMPUTE DIFFERENCE OF LOGS 
        #===========================================================================    
        hus_anom = MV.log(q1) - MV.log(q0)

        # KERNELS REQUIRE DETRENDED QV DIVIDED BY DELQ:
        #=========================================
        norm_hus_anom = np.ma.true_divide(hus_anom,Delq)
        norm_hus_anom.setAxisList(q1.getAxisList())    
        del hus_anom, Delq


    del avgta, qsl, qsl_plus1, qsi, qsi_plus1, qs, qs_plus1, wsl, wsl_plus1, wsi, wsi_plus1, q1, q0    
    gc.collect() 

    return norm_hus_anom


###########################################################################
def tile_uneven(data,data_to_match):
    """extend data to match size of data_to_match even if not a multiple of 12"""
    
    A12=len(data_to_match)/12
    ind=np.arange(12,)
    rep_ind = np.tile(ind,(A12+1))[:len(data_to_match)]
    
    rep_data = MV.array(np.array(data)[rep_ind]) 
    
    return rep_data
    
###########################################################################
def trapez_int_norm4(fx,trop_wts,atm_wts):
    """
    Same as previous, but uses the actual tropopause pressure rather 
    than the linearly varying tropopause used previously

    Perform trapezoidal-approximation integration of fx over the height 
    dimension (in pressure levels). fx dims expected to be [time,heights,lat,lon]
    and p should have dim = [heights] and should be surface-comes-first order.

    X-dim is normalized by 100 hPa to match Soden.  

    Mask out places where press is greater than surface press
    From Alex Jonko regarding weighting: "I use the difference of pressure level 
    midpoints. For the layer closest to the surface I use the difference between 
    surface pressure and the first midpoint above that. The lowest midpoint is 
    962.5 hPa, between 1000 hPa and 925 hPa. If ps > 962.5 hPa, then the first 
    layer thickness is ps - 962.5 hPa. If ps < 962.5 hPa, then the layer has 0 
    thickness, and you move up to the next midpoint." 
    """

    #TRAPEZOIDAL RULE INTEGRATION:
    #==============================
    #have 16 dp layers and 17 fx vals (ie fx vals are cell edges). Average each 2 
    #consecutive fx vals together to form midpt vals, then mult by dp for that layer.
    trop=(fx[:,:,0:-1,:]+fx[:,:,1:,:])/2. * trop_wts
    whole=(fx[:,:,0:-1,:]+fx[:,:,1:,:])/2. * atm_wts
    
    #trop_int=MV.sum(trop, axis=2) #sum over heights.
    #whole_int=MV.sum(whole, axis=2) #sum over heights.
    trop_int=np.sum(trop, axis=2) #sum over heights.
    whole_int=np.sum(whole, axis=2) #sum over heights.

    # Give back the axes attributes
    #trop_int.setAxisList([fx.getAxisList()[0],fx.getAxisList()[1],fx.getAxisList()[3],fx.getAxisList()[4]])
    #whole_int.setAxisList([fx.getAxisList()[0],fx.getAxisList()[1],fx.getAxisList()[3],fx.getAxisList()[4]])

    return (trop_int, whole_int)

